/*
     This file is part of libextractor.
     (C) 2002, 2003 Vidyut Samanta and Christian Grothoff

     libextractor is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     libextractor is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with libextractor; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
 **/

#include "extractor.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "getopt.h"
#define YES 1
#define NO 0

static void
printHelp ()
{
  printf ("USAGE: extract [OPTIONS] [FILENAME]*\n"
	  "-h, --help                : print this page\n"
	  "-v, --version             : print the version number\n"
	  "-l LIBRARY, --load=LIBRARY: load an extractor library\n"
	  "-n, --nodefault           : do not use the default set of extractors\n"
	  "-b, --bibtex              : Print bibtext output\n"
	  "-s, --split               : use keyword splitting\n"
	  "-f, --filename            : use the filename as a keyword\n"
	  "-p TYPE, --print=TYPE     : print only keywords of the given type\n"
	  "-x TYPE, --exclude=TYPE   : do not print keywords of the given type\n"
	  "-L, --list                : list all types\n"
	  "-V, --verbose             : be verbose\n"
	  "-d, --duplicates          : remove duplicates only if types match\n"
	  "-r, --remove-duplicates   : remove duplicates (disregard keyword type)\n"
	  "-a, --all                 : do not remove any duplicates\n",
	  "-B, --binary              : try the generic plaintext extractor\n");
}


/**
 * Print a keyword list to a file.
 * For debugging.
 * @param handle the file to write to (stdout, stderr), may NOT be NULL
 * @param keywords the list of keywords to print, may be NULL
 * @param print array indicating which types to print
 **/
static void
printSelectedKeywords (FILE * handle,
		       EXTRACTOR_KeywordList * keywords,
		       int *print, int verbose)
{
  while (keywords != NULL)
    {
      if (NULL == EXTRACTOR_getKeywordTypeAsString (keywords->keywordType))
	{
	  if (verbose == YES)
	    {
	      fprintf (handle, "INVALID TYPE - %s\n", keywords->keyword);
	    }
	}
      else if (print[keywords->keywordType] == YES)
	fprintf (handle,
		 "%s - %s\n",
		 EXTRACTOR_getKeywordTypeAsString (keywords->keywordType),
		 keywords->keyword);
      keywords = keywords->next;
    }
}

/**
 * Take title, auth, year and return a string
 **/
static char *
splice(char * title, char * auth, char * year)
{
  char * temp = (char*)malloc(sizeof(char)*16);
  int i = 0;
  bzero(temp, sizeof(char)*16);
  snprintf(temp, 15, "%.5s%.5s%.5s", auth, year, title);
  for ( i = 0; i < strlen(temp); i++ ) {
    if ( !isalnum(temp[i]) )
      temp[i] = '_';
    else
      temp[i] = tolower(temp[i]);
  }
  return temp;
}

/**
 * Print a keyword list in bibtex format to a file.
 * For debugging.
 * @param handle the file to write to (stdout, stderr), may NOT be NULL
 * @param keywords the list of keywords to print, may be NULL
 * @param print array indicating which types to print
 **/
/**
 * FIXME: We should generate the three letter abbrev of the month
 **/
static void
printSelectedKeywordsBibtex (FILE * handle,
		       EXTRACTOR_KeywordList * keywords,
		       int *print, const char * filename)
{
  char * last = NULL;
  if (print[keywords->keywordType] == YES)
    {
      char * title = NULL;
      char * author = NULL;
      char * note = NULL;
      char * date = NULL;
      char * publisher = NULL;
      char * organization = NULL;
      char * key = NULL;
      char * pages = NULL;
      char * year = NULL;
      char * month = NULL;

      title = EXTRACTOR_extractLastByString("title", keywords);
      if ( !title )
	title = EXTRACTOR_extractLastByString("filename", keywords);
      if ( !title )
	title = (char*)filename;
      last = title;

      author = EXTRACTOR_extractLastByString("author", keywords);
      if ( author )
	last = author;

      note = EXTRACTOR_extractLastByString("description", keywords);
      if ( !note )
	note = EXTRACTOR_extractLastByString("keywords", keywords);
      if ( !note )
	note = EXTRACTOR_extractLastByString("comment", keywords);
      if ( note )
	last = note;

      date = EXTRACTOR_extractLastByString("date", keywords);
      if ( !date )
	date = EXTRACTOR_extractLastByString("creation date", keywords);
      if ( date ) {
	if ( strlen(keywords->keyword) >= 7 ) {
	  year = (char*)malloc(sizeof(char)*5);
	  bzero(year, sizeof(char)*5);
	  month = (char*)malloc(sizeof(char)*3);
	  bzero(month, sizeof(char)*3);
	  year[0] = keywords->keyword[0];
	  year[1] = keywords->keyword[1];
	  year[2] = keywords->keyword[2];
	  year[3] = keywords->keyword[3];
	  month[0] = keywords->keyword[4];
	  month[1] = keywords->keyword[5];
	} else if ( strlen(keywords->keyword) >= 4 ) {
	  year = (char*)malloc(sizeof(char)*5);
	  bzero(year, sizeof(char)*5);
	  year[0] = keywords->keyword[0];
	  year[1] = keywords->keyword[1];
	  year[2] = keywords->keyword[2];
	  year[3] = keywords->keyword[3];
	}
      }
      if ( year )
	last = year;

      if ( month )
	last = month;

      publisher = EXTRACTOR_extractLastByString("publisher", keywords);
      if ( publisher )
	last = publisher;

      organization = EXTRACTOR_extractLastByString("organization", keywords);
      if ( organization )
	last = organization;

      key = EXTRACTOR_extractLastByString("subject", keywords);
      if ( key )
	last = key;

      pages = EXTRACTOR_extractLastByString("page count", keywords);
      if ( pages )
	last = pages;

      fprintf(handle, "@misc{ %s,\n",splice(title, author, year));
      if ( title )
	fprintf(handle, "    title = \"%s\"%s\n", title, 
	    (last == title)?"":",");
      if ( author )
	fprintf(handle, "    author = \"%s\"%s\n", author,
	    (last == author)?"":",");
      if ( note )
	fprintf(handle, "    note = \"%s\"%s\n", note,
	    (last == note)?"":",");
      if ( year )
	fprintf(handle, "    year = \"%s\"%s\n", year,
	    (last == year)?"":",");
      if ( month )
	fprintf(handle, "    month = \"%s\"%s\n", month,
	    (last == month)?"":",");
      if ( publisher )
	fprintf(handle, "    publisher = \"%s\"%s\n", publisher,
	    (last == publisher)?"":",");
      if ( organization )
	fprintf(handle, "    organization = \"%s\"%s\n", organization,
	    (last == organization)?"":",");
      if ( key )
	fprintf(handle, "    key = \"%s\"%s\n", key,
	    (last == key)?"":",");
      if ( pages )
	fprintf(handle, "    pages = \"%s\"%s\n", pages,
	    (last == pages)?"":",");

      fprintf(handle, "}\n\n");
    }
}

/**
 * Demo for libExtractor.
 * <p>
 * Invoke with a list of filenames to extract keywords
 * from (demo will use all the extractor libraries that
 * are available by default).
 **/
int
main (int argc, char *argv[])
{
  int i;
  EXTRACTOR_ExtractorList *extractors;
  EXTRACTOR_KeywordList *keywords;
  int option_index;
  int c;
  char *libraries = NULL;
  char *ldenv;
  int splitKeywords = NO;
  int verbose = NO;
  int useFilename = NO;
  int nodefault = NO;
  int *print;
  int defaultAll = YES;
  int duplicates = EXTRACTOR_DUPLICATES_REMOVE_UNKNOWN;
  int bibtex = NO;
  int binary = NO;

  print = malloc (sizeof (int) * EXTRACTOR_getHighestKeywordTypeNumber ());
  for (i = 0; i < EXTRACTOR_getHighestKeywordTypeNumber (); i++)
    print[i] = YES;		/* default: print everything */

  while (1)
    {
      static struct option long_options[] = {
	{"version", 0, 0, 'v'},
	{"help", 0, 0, 'h'},
	{"library", 1, 0, 'l'},
	{"bibtex", 0, 0, 'b'},
	{"nodefault", 0, 0, 'n'},
	{"split", 0, 0, 's'},
	{"filename", 0, 0, 'f'},
	{"print", 1, 0, 'p'},
	{"exclude", 1, 0, 'x'},
	{"list", 0, 0, 'L'},
	{"verbose", 0, 0, 'v'},
	{"duplicates", 0, 0, 'd'},
	{"remove-duplicates", 0, 0, 'r'},
	{"all", 0, 0, 'a'},
	{"binary", 0, 0, 'B'},
	{0, 0, 0, 0}
      };
      option_index = 0;
      c = getopt_long (argc,
		       argv, "vhbl:nsfp:x:LVdraB", 
		       long_options, 
		       &option_index);

      if (c == -1)
	break;			// No more flags to process
      switch (c)
	{
	case 'v':
	  printf ("Version %s\n", PACKAGE_VERSION);
	  return 0;
	case 'h':
	  printHelp ();
	  return 0;
	case 'b':
	  bibtex = YES;
	  break;
	case 'l':
	  if (optarg == NULL) {
	    fprintf(stderr,
		    "You must specify an argument for the -l option (option ignored).\n");
	    break;
	  }
	  libraries = optarg;
	  break;
	case 'L':
	  i = 0;
	  while (NULL != EXTRACTOR_getKeywordTypeAsString (i))
	    printf ("%s\n", EXTRACTOR_getKeywordTypeAsString (i++));
	  return 0;
	case 'p':
	  if (optarg == NULL) {
	    fprintf(stderr,
		    "You must specify an argument for the -p option (option ignored).\n");
	    break;
	  }
	  if (defaultAll == YES)
	    {
	      defaultAll = NO;
	      i = 0;
	      while (NULL != EXTRACTOR_getKeywordTypeAsString (i))
		print[i++] = NO;
	    }
	  i = 0;
	  while (NULL != EXTRACTOR_getKeywordTypeAsString (i))
	    {
	      if (0 == strcmp (optarg, EXTRACTOR_getKeywordTypeAsString (i)))
		{
		  print[i] = YES;
		  break;
		}
	      i++;
	    }
	  if (NULL == EXTRACTOR_getKeywordTypeAsString (i))
	    {
	      fprintf (stderr,
		       "Unknown keyword type %s, use -L to get a list.\n",
		       optarg);
	      return -1;
	    }
	  break;
	case 'B':
	  binary = YES;
	  break;
	case 'x':
	  if (optarg == NULL) {
	    fprintf(stderr,
		    "You must specify an argument for the -x option (option ignored).\n");
	    break;
	  }
	  i = 0;
	  while (NULL != EXTRACTOR_getKeywordTypeAsString (i))
	    {
	      if (0 == strcmp (optarg, EXTRACTOR_getKeywordTypeAsString (i)))
		{
		  print[i] = NO;
		  break;
		}
	      i++;
	    }
	  if (NULL == EXTRACTOR_getKeywordTypeAsString (i))
	    {
	      fprintf (stderr,
		       "Unknown keyword type %s, use -L to get a list.\n",
		       optarg);
	      return -1;
	    }
	  break;
	case 's':
	  splitKeywords = YES;
	  break;
	case 'f':
	  useFilename = YES;
	  break;
	case 'V':
	  verbose = YES;
	  break;
	case 'd':
	  duplicates = 0;
	  break;
	case 'a':
	  duplicates = -1;
	  break;
	case 'r':
	  duplicates = EXTRACTOR_DUPLICATES_TYPELESS;
	  break;
	case 'n':
	  nodefault = YES;
	  break;
	default:
	  fprintf (stderr,
		   "Unknown option %c. Aborting.\nUse --help to get a list of options.\n",
		   c);
	  return -1;
	}			/* end of parsing commandline */
    }				/* while (1) */

  if (argc - optind < 1)
    {
      fprintf (stderr,
	       "Invoke with list of filenames to extract keywords form!\n");
      return -1;
    }

  /* build list of libraries */
  if (nodefault == NO)
    extractors = EXTRACTOR_loadDefaultLibraries ();
  else
    extractors = NULL;
  if (useFilename == YES)
    extractors = EXTRACTOR_addLibrary (extractors,
				       "libextractor_filename");
  if (libraries != NULL)
    extractors = EXTRACTOR_loadConfigLibraries (extractors, libraries);

  if (binary == YES)
      extractors = EXTRACTOR_addLibraryLast(extractors,
					    "libextractor_printable");
      
  if (splitKeywords == YES)
    extractors = EXTRACTOR_addLibraryLast(extractors,
					  "libextractor_split");


  if (verbose == YES)
    {
      /* print list of all used extractors */
    }
  /* extract keywords */
  if ( bibtex == YES )
    fprintf(stdout, "% BiBTeX file\n");
  for (i = optind; i < argc; i++)
    {
      keywords = EXTRACTOR_getKeywords (extractors, argv[i]);
      if (duplicates != -1 || bibtex == YES)
	keywords = EXTRACTOR_removeDuplicateKeywords (keywords, duplicates);
      if (verbose == YES && bibtex == NO)
	printf ("Keywords for file %s:\n", argv[i]);
      if (bibtex == YES)
	printSelectedKeywordsBibtex (stdout, keywords, print, argv[i]);
      else
	printSelectedKeywords (stdout, keywords, print, verbose);
      if (verbose == YES && bibtex == NO)
	printf ("\n");
      EXTRACTOR_freeKeywords (keywords);
    }
  free (print);
  EXTRACTOR_removeAll (extractors);

  return 0;
}
