/*
     This file is part of libextractor.
     (C) 2002, 2003 Vidyut Samanta and Christian Grothoff

     libextractor is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     libextractor is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with libextractor; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
 **/

#include "extractor.h"
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>

#define DEBUG_EXTRACT_HTML 1

#include "libhtmlparse.h"

/**
 * 0: ignore, 1: add keyword
 **/
static int nextTextAction = 0;

/**
 * If nextTextAction == 1, this gives the type of the
 * keyword.
 **/
static EXTRACTOR_KeywordType nextKeywordType;

/**
 * Result of the current pass.
 **/
static struct EXTRACTOR_Keywords * result;

/**
 * Add a keyword.
 **/
static struct EXTRACTOR_Keywords * addKeyword(EXTRACTOR_KeywordType type,
					      char * keyword,
					      struct EXTRACTOR_Keywords * next) {
  EXTRACTOR_KeywordList * result;

  if (keyword == NULL)
    return next;
  result = (EXTRACTOR_KeywordList*)malloc(sizeof(EXTRACTOR_KeywordList));
  result->next = next;    
  result->keyword = strdup(keyword);
  result->keywordType = type;
  return result;
}

/**
 * Called by the parser whenever we see text.
 **/
static int texts (char *comment) {
  if (nextTextAction) {
    result = addKeyword(nextKeywordType,
			comment,
			result);
    nextTextAction = 0;
  }
  return 0;
}

static int hasTag(char * arg,
		  char * val,
		  struct ArgvTable * args,
		  int numargs) {
  int i;
  for (i=0;i<numargs;i++) 
    if (0 == strcasecmp(args[i].arg, arg))
      if (0 == strcasecmp(args[i].val, val))
	return 1;
  return 0;
}

static char * getTag(char * arg,
		     struct ArgvTable * args,
		     int numargs) {
  int i;
  for (i=0;i<numargs;i++) 
    if (0 == strcasecmp(args[i].arg, arg))
      return args[i].val;
  return NULL;
}

static struct {
  char * name;
  EXTRACTOR_KeywordType type;
} tagmap[] = {
   { "author" , EXTRACTOR_AUTHOR}, 
   { "description" , EXTRACTOR_DESCRIPTION},
   { "language", EXTRACTOR_LANGUAGE},
   { "rights", EXTRACTOR_COPYRIGHT}, 
   { "publisher", EXTRACTOR_PUBLISHER},
   { "date", EXTRACTOR_DATE},
   { "keywords", EXTRACTOR_KEYWORDS},
   {NULL, EXTRACTOR_UNKNOWN},
};



static int starttag (char *tag, 
		     struct ArgvTable *args, 
		     int numargs) {
  int i;
  
  if (0 == strcasecmp(tag,"title")) {
    nextTextAction = 1;
    nextKeywordType = EXTRACTOR_TITLE;
    return 0;
  } 
  if (0 == strcasecmp(tag,"meta")) {
    i = 0;
    while (tagmap[i].name != NULL) {
      if (hasTag("name",tagmap[i].name,args, numargs)) 
	result = addKeyword(tagmap[i].type,
			    getTag("content",
				   args, numargs),
			    result);
      i++;
    }
  }
  /*
    Don't do this, you can't be certain...*/
#if I_AM_CERTAIN
  if (0 == strcasecmp(tag,"html")) {
    result = addKeyword(EXTRACTOR_MIMETYPE,
			"text/html",
			result);
    return 0;
  } 
#endif  
  return 0;
}

int endtag (char *tag) {
  nextTextAction = 0;
  return 0;
}


/* mimetype = text/html */
struct EXTRACTOR_Keywords * extract(const char * filename,
				    char * data,
				    const size_t size,
				    struct EXTRACTOR_Keywords * prev) {
  char backup;

  result = prev;
  backup = data[size-1];
  data[size-1] = 0;
  textCallBack  = texts;
  startCallBack = starttag;
  /* the parser requires 0-termination. We just
     overwrite the last character in data and
     restore it later, assuming that it can
     hardly be a keyword in a valid HTML text... 
  */
  parse(data);
  data[size-1] = backup;
  return result;
}

#define HAVE_MAIN 0
#if HAVE_MAIN
int main (int argc, char **argv) {
  int file;
  char * buffer;
  struct stat fstatbuf;
  size_t size;

  if (argc != 2) {
    fprintf(stderr, 
	    "Call with filename as argument\n");
    return -1;
  }  
  file = open(filename,O_RDONLY);
  if (-1 == file) 
    return NULL;
  if (-1 == fstat(file, &fstatbuf)) {
    close(file);
    return NULL;
  }
  size = fstatbuf.st_size;
  buffer = mmap(NULL, size, PROT_READ, MAP_SHARED, file, 0);  
  close(file);

  EXTRACTOR_printKeywords(stdout,
			  extract(argv[1], 
				  NULL, 
				  buffer,
				  size,
				  NULL));
  free(size);
  return 0;
}
#endif
