/*
     This file is part of libextractor.
     (C) 2002, 2003 Vidyut Samanta and Christian Grothoff

     libextractor is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     libextractor is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with libextractor; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
 **/

/**
 * Todo:
 *
 * - add linguistic support (for non-dictionary based mode)
 * - use word-counting to weigh results if we get too many
 * - improve heuristics
 **/

#include "config.h"
#include "extractor.h"
#if HAVE_ASPELL_H
#include <aspell.h>
#define SPELLER AspellSpeller *
#else
#if HAVE_PSPELL_PSPELL_H
#include <pspell/pspell.h>
#define SPELLER PspellManager *
#else
#warning Neither aspell nor pspell headers installed. Will run without spell-checker.
#define SPELLER void *
#endif
#endif



#include <string.h>
#include <stdlib.h>
#include <ctype.h>

static int MINIMUM_KEYWORD_LENGTH = 5;

static struct EXTRACTOR_Keywords * addKeyword(struct EXTRACTOR_Keywords * list,
					      char * keyword) {
  EXTRACTOR_KeywordList * next;
  next = malloc(sizeof(EXTRACTOR_KeywordList));
  next->next = list;    
  next->keyword = keyword;
  next->keywordType = EXTRACTOR_UNKNOWN;
  return next;
}

/* test if the keyword is only or mostly numbers 
   or whitespace (reason to reject) */
static int numberTest(char * keyword) {
  unsigned int cnt;
  unsigned int hex;
  unsigned int space;
  char * pos;
  
  cnt = 0;
  hex = 0;
  space = 0;
  pos = keyword;
  while (0 != *pos) {
    if (isdigit(*pos))
      cnt++;
    if (isspace(*pos))
      space++;
    if (isxdigit(*pos))
      hex++;
    pos++;    
  }
  if ( ((cnt+space)*2 > pos - keyword) ||
       ((hex+space)*4 > (pos - keyword) * 3) ) {
    /* more than 50% numbers or whitespace or
       more than 75% hex or whitespace */
    return -1;
  } else
    return 0;
}

/**
 * count the number of words in the given keyword, and if it is over
 * 50%, accept the whole thing, otherwise reject (both proportional
 * to word length).
 **/
static int tokenTest(SPELLER spell_checker,
		     char * keyword) {
  int tokenCount;
  int goodWords;
  char * start;

  tokenCount = 1;
  goodWords = strlen(keyword);
  start = keyword;
  while (*keyword != 0) {
    if (isalpha(*keyword)) {
      keyword++;
      continue;
    }
    if (isdigit(*keyword)) {
      int isMixed;
      
      isMixed = 0;
      if (start != keyword) 
	isMixed = 2;
      /* only count mixed word-letter combinations
	 as bad, but these with a weight of 2 */
      tokenCount += isMixed * (keyword - start); 
      start = keyword;
      continue;
    }
    keyword++;
  }
  return (goodWords * 2 - tokenCount);
}

/**
 * to be fixed by Linguists 
 * @param word (alphabetic characters without spaces)
 * @return 0 if it is no word, 1 if it is
 **/
static int wordTest(SPELLER spell_checker,
		    char * word) {
  int i;
  if (strlen(word) <= 2)
    return 0;
  for (i=0;i<strlen(word);i++) {
    if (word[i] == ' ') {
      /* ok, recurse into substrings! */
      word[i] = '\0';
      return wordTest(spell_checker,
		      word) &&
	wordTest(spell_checker,
		 &word[i+1]);
    }
  }

  if (spell_checker == NULL) {
    /* Linguists may want to fixme here! 
     -- we may also want to call this code once it exists
     even if we have aspell; think of a keyword like "monkeyboy" which
     aspell will never have but which could be used... */
    if (strlen(word) < MINIMUM_KEYWORD_LENGTH)
      return 0;
    if (numberTest(word) < 0) 
      return 0;
    return 1;
  }
#if HAVE_ASPELL_H
  return aspell_speller_check(spell_checker, word, strlen(word));
#else
#if HAVE_PSPELL_PSPELL_H
  return pspell_manager_check(spell_checker, word, strlen(word));
#else
  return 0; /* should never happen */
#endif
#endif
}


static void testKeyword(SPELLER spell_checker,
			size_t * last,
			size_t pos,
			char * data,
			struct EXTRACTOR_Keywords ** head,
			struct EXTRACTOR_Keywords ** tail) {
  char * keyword;
  char * alternate;
  int score;
  int i;
  int p;
  int len;
  int w;
  int dpos;

  keyword = malloc(pos - *last + 1);
  alternate = malloc(pos - *last + 1);
  memcpy(keyword, 
	 &data[*last],
	 pos-*last);
  keyword[pos-*last] = '\0';

  /* remove whitespace */
  dpos = *last;
  p=0;
  len = strlen(keyword);
  for (i=0;i<len;i++) {
    if ( (p==0) && (isspace(keyword[i])))
      continue;
    if (isspace(keyword[i]))
      if (isspace(keyword[p])) {
	continue;
      }
    if (isspace(keyword[i]))
      keyword[i] = ' '; /* repalce tab/newline by space */
    keyword[p++] = keyword[i];
  }
  /* move/copy zero-termination */
  if (p == 0) {
    free(keyword);
    free(alternate);
    *last += len;
    return;
  }
    
  if (isspace(keyword[p-1])) {
    keyword[p-1] = '\0';
  } else
    keyword[p++] = '\0';
  p = 0;
  w = 0 ;
  len = strlen(keyword);
  for (i=0;i<len;i++) {
    if (isalnum(keyword[i])) {
      alternate[w++] = keyword[i];
      continue;
    }
    if ( (keyword[i] == ' ') || 
	 (keyword[i] == '\0') ) {
      keyword[i] = '\0';
      alternate[w] = '\0';
      if (wordTest(spell_checker, 
		   &keyword[p])) {	
	int k;
	for (k=0;k<strlen(&keyword[p]);k++) {
	  while (data[dpos] != keyword[p+k])
	    dpos++;
	  dpos++;
	}
	if (*tail != NULL) {
	  (*tail)->next = addKeyword(NULL, strdup(&keyword[p]));
	  *tail = (*tail)->next;
	} else { 
	  *tail = addKeyword(NULL, strdup(&keyword[p]));
	  *head = *tail;
	}
      } else
	if (wordTest(spell_checker, 
		     alternate)) {
	  int k;
	  for (k=0;k<strlen(alternate);k++) {
	    while (data[dpos] != alternate[k])
	      dpos++;
	    dpos++;
	  }
	  if (*tail != NULL) {
	    (*tail)->next = addKeyword(NULL, strdup(alternate));      
	    *tail = (*tail)->next;
	  } else {
	    *tail = addKeyword(NULL, strdup(alternate));      
	    *head = *tail;
	  }
	}
      p = i+1;
      w = 0;		   
    }          
    if (!isprint(keyword[i]) )
      keyword[i] = ' ';
  }
  if (pos - dpos > 20)
    dpos = pos - 20; /* linearize complexity! Without this,
			we're quadratic, with this, it's O(n) with
			a constant factor of 20.  20 means that we
			may miss keywords that are LONGER than 20
			characters. */
  *last = dpos+1;
  free(keyword);
  free(alternate);
}
	
/* which mime-types should not be subjected to
   the printable extractor (e.g. because we have
   a more specific extractor available)? */
static char * blacklist[] = {
  "image/jpeg",
  "image/gif",
  "image/png",
  "image/x-png",
  "video/real",
  "audio/real",
  "application/x-rpm",
  "audio/mp3",
  "audio/mp2",
  "audio/mp1",
  "application/x-ogg",
  "video/asf",
  "video/quicktime",
  NULL,
};

/* "man strings" gives for now an adequate description of
   what we are doing here. */
struct EXTRACTOR_Keywords * extract(char * filename,
				    char * data,
				    size_t size,
				    struct EXTRACTOR_Keywords * prev) {
  size_t last;
  size_t pos;
  SPELLER spell_checker; 
  char * lang;

#if HAVE_ASPELL_H
  AspellConfig * spell_config; 
  AspellCanHaveError * possible_err; 
  AspellError * definite_err;
#else
#if HAVE_PSPELL_PSPELL_H
  PspellConfig * spell_config;
  PspellCanHaveError * possible_err;
#else
#endif
#endif
  char * mime;
  struct EXTRACTOR_Keywords * head = NULL;
  struct EXTRACTOR_Keywords * tail = NULL;

  /* if the mime-type of the file is blacklisted, don't
     run the printable extactor! */
  mime = EXTRACTOR_extractLast(EXTRACTOR_MIMETYPE,
			       prev);
  if (mime != NULL) {
    int j;
    j = 0;
    while (blacklist[j] != NULL) {
      if (0 == strcmp(blacklist[j], mime))
	return prev;
      j++;
    }
  }  
  
  lang = getenv("LANG");
  if ( (lang == NULL) || 
       (strlen(lang) == 0) || 
       (0 == strcmp(lang, "C") ) )
    lang = "en_US";
#if HAVE_ASPELL_H
  spell_config = new_aspell_config(); 
  
  aspell_config_replace(spell_config, "lang", lang); 
  possible_err = new_aspell_speller(spell_config); 
  if (aspell_error(possible_err) != NULL) {
    fprintf(stderr,
	    "ERROR: aspell speller could not be created (%s) for language >>%s<<\n",
	    aspell_error_message(possible_err),
	    lang); 
    spell_checker = NULL;
  } else 
    spell_checker = to_aspell_speller(possible_err);   

#else
#if HAVE_PSPELL_PSPELL_H
  spell_config = new_pspell_config();
  pspell_config_replace(spell_config, "language-tag", lang);
  possible_err = new_pspell_manager(spell_config);
  delete_pspell_config(spell_config); 
  if (pspell_error_number(possible_err) != 0) {
    fprintf(stderr,
	    "ERROR: pspell speller could not be created (%s) for language >>%s<<\n",
	    pspell_error_message(possible_err),
	    lang); 
    spell_checker = NULL;
  } else 
    spell_checker = to_pspell_manager(possible_err);
#else
  spell_checker = NULL;
#endif
#endif
 

  last = 0;
  pos = 0;
  while (pos < size) {
    if ( isalnum(data[pos]) 
	 || isblank(data[pos]) ) {
      pos++;
      continue;
    }   
    testKeyword(spell_checker, 
		&last, 
		pos, 
		data,
		&head,
		&tail);
    pos++;
  }
  testKeyword(spell_checker, 
	      &last,
	      pos, 
	      data,
	      &head,
	      &tail);
  if (tail != NULL) {
    tail->next = prev;
    prev = head;
  }


#if HAVE_ASPELL_H
  delete_aspell_speller(spell_checker); 
  delete_aspell_config(spell_config); 
#else
#if HAVE_PSPELL_PSPELL_H
  delete_pspell_manager(spell_checker); 
#endif
#endif

  return prev;
}


/* end of printableextractor.c */
